/* encache.c - Generic code for setting up pagetables or the protection unit,
 * and enabling the cache. */

/* Copyright (C) 1997 Advanced RISC Machines Limited.
 * All Rights Reserved. */

/* This file refers to the A.R.M.--The ARM Architecture Reference Manual */

/* BlankPageTable defines whether 'pagetab.s' exports an empty table,
 * or a completed page table. There is an identical define in 'pagetab.s',
 * and the two should be kept in step.
 */
#ifndef BlankPageTable
#define BlankPageTable 0
#endif

/* The CP15 "ID" register (c0) contains the processor name and revision.
 * See A.R.M. section 7.4.1 */
__inline unsigned long readCP15R0(void)
{
  unsigned long id;
  __asm { MRC P15, 0, id, c0, c0; }
  return id;
}

/* The pagetable has to be declared externally -
 * it needs to be in an assembler file so it can be aligned
 * to a 16kB boundary.
 * Note: extern unsigned long *pagetable; is NOT the same
 * thing (see ANSI C standard), and will not work in this
 * instance.
 */
extern unsigned long pagetable[4096];

#if BlankPageTable
/*
 * pagetab.s exports a blank page table - use code to
 * fill it in.
 */
enum access {
  NO_ACCESS,
  NO_USR_W,
  SVC_RW,
  ALL_ACCESS
};

enum entrytype {
  INVALID,
  PAGE,
  SECTION
};

#define U_BIT 16
#define C_BIT 8
#define B_BIT 4

#define L1Entry(type,addr,dom,ucb,acc) \
  ( (type == SECTION) ? ( ((addr) & 0xfff00000) |        \
                          ((acc) << 10) | ((dom) << 5) | \
                          (ucb) | (type) ) :             \
    (type == PAGE) ? ( ((addr) &0xfffffc00) |            \
                       ((dom) << 5) |                    \
                       ((ucb) & U_BIT) | (type) ) :      \
    0)

static unsigned long const *createPageTable(void)
{
  int i;
  unsigned long *p = pagetable;
  unsigned long entry;

  /* See A.R.M. section 7.6 */
  /* first region - 16MB of cacheable/bufferable */
  entry = L1Entry(SECTION, 0, 0,
                  U_BIT|C_BIT|B_BIT, ALL_ACCESS);
  for (i = 0; i < 16; i++) {
    *p++ = entry | (i << 20);
  }

  /* second region - upto 1MB short of 2GB,
   * not cacheable/not bufferable */
  entry = L1Entry(SECTION, 0, 0, 0, ALL_ACCESS);
  for (; i < 2048 - 1; i++) {
    *p++ = entry | (i << 20);
  }

  /* third region - up to 2GB cacheable/bufferable */
  entry = L1Entry(SECTION, 0, 0,
                  U_BIT|C_BIT|B_BIT, ALL_ACCESS);
  for (; i < 2048; i++) {
    *p++ = entry | (i << 20);
  }

  /* rest of memory - not cacheable/not bufferable */
  entry = L1Entry(SECTION, 0, 0, 0, ALL_ACCESS);
  for (; i<4096; i++) {
    *p++ = entry | (i << 20);
  }

  return pagetable;
}

#else  /* BlankPageTable */

/* pagetab.s exports a complete page table */
#define createPageTable() ((unsigned long const *)pagetable)

#endif

/* These are all the bits currently defined for the control register */
/* A.R.M. 7.4.2 */
#define MMU_V 0x2000
#define MMU_I 0x1000
#define MMU_Z 0x0800
#define MMU_F 0x0400
#define MMU_R 0x0200
#define MMU_S 0x0100
#define MMU_B 0x0080
#define MMU_L 0x0040
#define MMU_D 0x0020
#define MMU_P 0x0010
#define MMU_W 0x0008
#define MMU_C 0x0004
#define MMU_A 0x0002
#define MMU_M 0x0001

/* Some operations are different between architecture v.3 and v.4 */
typedef enum {
  arch_3,
  arch_4
} architecture;

typedef enum { FALSE, TRUE } bool;

/*
 * The functions below take arguments to specify which "caches" the
 * action is to be directed at. For the I-cache, pass "MMU_I". For
 * the D-cache, "MMU_C". For both, pass "MMU_ID". For combined ID-Cache
 * processors, use "MMU_C"
 */
#define MMU_ID (MMU_I + MMU_C)

/*
 * Inline functions for MMU functions
 */

/* Set the page tables base register: register 2 (A.R.M. 7.4.3) */
__inline void mmuSetPageTabBase(unsigned long const *pagetab)
{
  __asm { MCR  P15, 0, pagetab, c2, c0; }
}

/* Set the domain access-control register: register 3 (A.R.M. 7.4.4) */
__inline void mmuSetDomainAccessControl(unsigned long flags)
{
  __asm { MCR  P15, 0, flags, c3, c0; }
}

/* Flush the cache(s).
 * Architectures 3 and 4 differs in how to flush the cache -- this
 * code handles both
 */
__inline void mmuFlushCache(architecture arch, unsigned caches)
{
  unsigned long dummy;

  switch (arch) {
  case arch_3:
    /* no architecture 3 processor has a split cache */
    /* flush cache: write any value to register 7 (A.R.M. 7.5.8) */
    __asm { MCR  P15, 0, dummy, c7, c0, 0 }
    break;

  case arch_4:
  default:
    /* flush cache(s): write to register 7, with flags (A.R.M. 7.4.8) */
    switch (caches) {
    case MMU_C:
      __asm { MCR  P15, 0, dummy, c7, c6, 0 }
      break;
    case MMU_I:
      __asm { MCR  P15, 0, dummy, c7, c5, 0 }
      break;
    case MMU_ID:
      __asm { MCR  P15, 0, dummy, c7, c7, 0 }
      break;
    }
    break;
  }
}

/* Flush the TLB(s)
 * Again, architectures 3 and 4 differ, so this code handles both
 */
__inline void mmuFlushTLB(unsigned arch, unsigned tlbs)
{
  unsigned long dummy;
  switch (arch) {
  case arch_3:
    /* no architecture 3 processor has a split MMU */
    /* flush TLB: write any value to register 4 (A.R.M. 7.5.6) */
    __asm { MCR  P15, 0, dummy, c5, c0; }
    break;

  case arch_4:
  default:
    /* flush TLB(s): write to register 8, with flags (A.R.M. 7.4.9) */
    switch (tlbs) {
    case MMU_C:
      __asm { MCR  P15, 0, dummy, c8, c6, 0; }
      break;
    case MMU_I:
      __asm { MCR  P15, 0, dummy, c8, c5, 0; }
      break;
    case MMU_ID:
      __asm { MCR  P15, 0, dummy, c8, c7, 0; }
      break;
    }
    break;
  }
}

/* Enable the cache/MMU/TLB etc.
 * See A.R.M. section 7.4.2
 */
__inline void enable(unsigned long flags)
{
  __asm { MCR P15, 0, flags, c1, c0; }
}


/* StrongARM clock switching */
/* To allow the core to run at full speed, the StrongARM needs
 *"clock-switching" enabled. See SA-110 Technical Reference Manual,
 * Section 8, and section 5.2.11.
 */
__inline void saClockSwitch(bool enable)
{
  unsigned long dummy;
  if (enable) {
    __asm { MCR  P15, 0, dummy, c15, c1, 2; }
  } else {
    __asm { MCR  P15, 0, dummy, c15, c2, 2; }
  }
}

/*
 * Protection unit
 *
 * For details, see ARM940 or ARM740T data sheet.
 */

/* Regions can only be one of a number of (power-of-two) sizes */
typedef enum {
  size_4k=11,size_8k,   size_16k,  size_32k,
  size_64k,  size_128k, size_256k, size_512k,
  size_1M,   size_2M,   size_4M,   size_8M,
  size_16M,  size_32M,  size_64M,  size_128M,
  size_256M, size_512M, size_1G,   size_2G,
  size_4G
} regionSize;

/* ARM940 has control over both caches individually.
 * The functions we define will do either, depending on the bits
 * set in the 'caches' argument.
 *
 * Because of the way inline assembler works--it needs all the fields
 * to be real constants, rather than variables which happen to be constant--
 * we need one "WriteRegion" function per Region. This code only uses
 * 3 regions, so we don't define all 8 functions.
 */
__inline void puWriteRegion0(
    unsigned long base, regionSize size, bool enable, unsigned caches)
{
  if (caches & MMU_I)
    __asm { MCR  P15, 0, base + (size << 1) + enable, c6, c0, 1; }
  if (caches & MMU_C)
    __asm { MCR  P15, 0, base + (size << 1) + enable, c6, c0, 0; }
}

__inline void puWriteRegion1(
    unsigned long base, regionSize size, bool enable, unsigned caches)
{
  if (caches & MMU_I)
    __asm { MCR  P15, 0, base + (size << 1) + enable, c6, c1, 1; }
  if (caches & MMU_C)
    __asm { MCR  P15, 0, base + (size << 1) + enable, c6, c1, 0; }
}

__inline void puWriteRegion2(
    unsigned long base, regionSize size, bool enable, unsigned caches)
{
  if (caches & MMU_I)
    __asm { MCR  P15, 0, base + (size << 1) + enable, c6, c2, 1; }
  if (caches & MMU_C)
    __asm { MCR  P15, 0, base + (size << 1) + enable, c6, c2, 0; }
}

/* Functions to write to the region control registers--c2, c3 and c5 */
__inline void puSetCacheable(unsigned long regions, unsigned caches)
{
  /* mask the value written to 8 bits (a valid ARM constant) */
  if (caches & MMU_I)
    __asm { MCR  P15, 0, regions & 0xffUL, c2, c0, 1 }
  if (caches & MMU_C)
    __asm { MCR  P15, 0, regions & 0xffUL, c2, c0, 0 }
}

__inline void puSetBufferable(unsigned long regions, unsigned caches)
{
  if (caches & MMU_I)
    __asm { MCR P15, 0, regions & 0xffUL, c3, c0, 1 }
  if (caches & MMU_C)
    __asm { MCR  P15, 0, regions & 0xffUL, c3, c0, 0 }
}

__inline void puSetProtection(unsigned long regions, unsigned caches)
{
  /* don't mask the value written; converting 0xffffffffL to 0xffffL
   * forces the compiler to insert an extra instruction */
  if (caches & MMU_I)
    __asm { MCR  P15, 0, regions, c5, c0, 1 }
  if (caches & MMU_C)
    __asm { MCR  P15, 0, regions, c5, c0, 0 }
}

/* Flushing the cache on a protection unit processor is done in the same
 * way as for an architecture 4 MMU
 */
__inline void puFlushCache(unsigned caches)
{
  mmuFlushCache(arch_4, caches); /* same as for arch_4 */
}

unsigned long initMMU(void)
{
  unsigned long id;

  /* Determine the processor from bits 15:4 of the "ID" register */
  id = readCP15R0();

  switch (id & 0xff00UL) {
  case 0x0600:                      /* ARM6xx */
  case 0x7000:                      /* ARM70x */
  case 0x7100:                      /* ARM71x */
    /* architecture v. 3 */
    /* set up the page table, domain control registers */
    mmuSetPageTabBase(createPageTable());
    mmuSetDomainAccessControl(3);
    /* be sure to flush cache and TLB before enabling them */
    mmuFlushCache(arch_3, MMU_C);
    mmuFlushTLB(arch_3, MMU_C);
    /* write to control register :-
     * 32-bit data and program space,
     * write-buffer on, cache on, MMU on
     */
    enable(MMU_D+MMU_P+MMU_W+MMU_C+MMU_M);
    break;

  case 0x8100:                      /* ARM810 */
    /* architecture v. 4 with branch prediction */
    /* set up the page table, domain control registers */
    mmuSetPageTabBase(createPageTable());
    mmuSetDomainAccessControl(3);
    /* flush cache and TLB */
    mmuFlushCache(arch_4, MMU_C);
    mmuFlushTLB(arch_4, MMU_C);
    /* write to control register :-
     * branch prediction on,
     * 32-bit data and program space,
     * write-buffer on, cache on, MMU on
     */
    enable(MMU_Z+MMU_D+MMU_P+MMU_W+MMU_C+MMU_M);
    break;

  case 0xa100:                      /* StrongARM SA-110 */
    /* architecture v. 4 with I cache */
    /* enable I cache - doesn't need MMU enabling */
    enable(MMU_I+MMU_D+MMU_P);
    /* StrongARM also needs "clock switching" enabled
     * for full performance
     */
    saClockSwitch(TRUE);
    /* set up the page table, domain control registers */
    mmuSetPageTabBase(createPageTable());
    mmuSetDomainAccessControl(3);
    /* flush caches and TLBs */
    mmuFlushCache(arch_4, MMU_ID);
    mmuFlushTLB(arch_4, MMU_ID);
    /* write to control register :-
     * I-cache on, 32-bit data and program space,
     * write-buffer on, D-cache on, MMU on
     */
    enable(MMU_I+MMU_D+MMU_P+MMU_W+MMU_C+MMU_M);
    break;

  case 0x7400:			    /* ARM740 */
    /* protection unit */
    /* setup regions */
    /* region 0 will cover the whole memory, then two
     * regions are used to cover the cacheable parts. Mark
     * regions for both Instructions and Data */
    puSetCacheable(~1ul, MMU_C);    /* not region 0 (data cache only) */
    puSetBufferable(~1ul, MMU_C);   /* not region 0 */
    puSetProtection(~0ul, MMU_C);   /* all regions */
    /* background region - not cacheable */
    puWriteRegion0(0x0, size_4G, TRUE, MMU_C);
    /* first region - 16MB */
    puWriteRegion1(0x0, size_16M, TRUE, MMU_C);
    /* second region - 1MB */
    puWriteRegion2((1ul << 31) - (1ul << 20), size_1M, TRUE, MMU_C);
    /* flush caches */
    puFlushCache(MMU_C);
    /* write to control register:
     * I-cache on, D-cache on, PU on
     * SBZ bits
     */
    enable(MMU_I+MMU_L+MMU_D+MMU_P+MMU_W+MMU_C+MMU_M);
    break;

  case 0x9400:                      /* ARM940 */
    /* protection unit */
    /* same as the ARM740, except we have two caches, so we have to do
     * everything twice -- i.e. we pass "MMU_ID" around in place of "MMU_C"
     */
    puSetCacheable(~1ul, MMU_ID);   /* instruction and data cache */
    puSetBufferable(~1ul, MMU_ID);
    puSetProtection(~0ul, MMU_ID);
    puWriteRegion0(0x0, size_4G, TRUE, MMU_ID);
    puWriteRegion1(0x0, size_16M, TRUE, MMU_ID);
    puWriteRegion2((1ul << 31) - (1ul << 20), size_1M, TRUE, MMU_ID);
    puFlushCache(MMU_ID);
    /* write to control register:
     * I-cache on, D-cache on, PU on
     * SBZ bits
     */
    enable(MMU_I+MMU_L+MMU_D+MMU_P+MMU_W+MMU_C+MMU_M);
    break;
  }
  return id;
}

/* end of file encache.c */
